/*
 *
 *    OPEN-XCHANGE legal information
 *
 *    All intellectual property rights in the Software are protected by
 *    international copyright laws.
 *
 *
 *    In some countries OX, OX Open-Xchange, open xchange and OXtender
 *    as well as the corresponding Logos OX Open-Xchange and OX are registered
 *    trademarks of the OX Software GmbH group of companies.
 *    The use of the Logos is not covered by the GNU General Public License.
 *    Instead, you are allowed to use these Logos according to the terms and
 *    conditions of the Creative Commons License, Version 2.5, Attribution,
 *    Non-commercial, ShareAlike, and the interpretation of the term
 *    Non-commercial applicable to the aforementioned license is published
 *    on the web site http://www.open-xchange.com/EN/legal/index.html.
 *
 *    Please make sure that third-party modules and libraries are used
 *    according to their respective licenses.
 *
 *    Any modifications to this package must retain all copyright notices
 *    of the original copyright holder(s) for the original code used.
 *
 *    After any such modifications, the original and derivative code shall remain
 *    under the copyright of the copyright holder(s) and/or original author(s)per
 *    the Attribution and Assignment Agreement that can be located at
 *    http://www.open-xchange.com/EN/developer/. The contributing author shall be
 *    given Attribution for the derivative code and a license granting use.
 *
 *     Copyright (C) 2016-2020 OX Software GmbH
 *     Mail: info@open-xchange.com
 *
 *
 *     This program is free software; you can redistribute it and/or modify it
 *     under the terms of the GNU General Public License, Version 2 as published
 *     by the Free Software Foundation.
 *
 *     This program is distributed in the hope that it will be useful, but
 *     WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *     or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
 *     for more details.
 *
 *     You should have received a copy of the GNU General Public License along
 *     with this program; if not, write to the Free Software Foundation, Inc., 59
 *     Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

package com.openexchange.guard.mail.service.impl.guests;

import com.openexchange.exception.OXException;
import com.openexchange.mail.api.IMailProperties;
import com.openexchange.mail.api.MailAccess;
import com.openexchange.mail.api.MailConfig;
import com.openexchange.mail.api.MailLogicTools;
import com.openexchange.session.Session;

/**
 * {@link GuardGuestMailAccess}
 *
 * @author <a href="mailto:benjamin.gruedelbach@open-xchange.com">Benjamin Gruedelbach</a>
 * @since v7.10.0
 */
public class GuardGuestMailAccess extends MailAccess<IGuardGuestFolderStorage, IGuardGuestMessageStorage> {

    private IGuardGuestFolderStorage folderStorage;
    private IGuardGuestMessageStorage messageStorage;
    private MailLogicTools mailLogicTools;
    private static IMailProperties DEFAULT_GUARD_GUEST_MAIL_PROPERTIES = new IMailProperties() {

        @Override
        public void waitForLoading() throws InterruptedException {
            //NO-OP
        }

        @Override
        public boolean isUserFlagsEnabled() {
            return false;
        }

        @Override
        public boolean isSupportSubscription() {
            return false;
        }

        @Override
        public boolean isIgnoreSubscription() {
            return true;
        }

        @Override
        public boolean isAllowNestedDefaultFolderOnAltNamespace() {
            return false;
        }

        @Override
        public int getMailFetchLimit() {
            return 1000;
        }

        @Override
        public boolean hideInlineImages() {
            //TODO: check for c.o.mail.hideInlineImages?
            return true;
        }
    };

    /**
     * Initializes a new {@link GuardGuestMailAccess}.
     * @param session
     */
    protected GuardGuestMailAccess(Session session) {
        super(session);
    }

    /**
     * Initializes a new {@link GuardGuestMailAccess}.
     * @param session
     * @param accountId
     */
    public GuardGuestMailAccess(Session session, int accountId) {
        super(session, accountId);
    }

    /**
     * serialVersionUID
     */
    private static final long serialVersionUID = -6899772477003364633L;



    /* (non-Javadoc)
     * @see com.openexchange.mail.api.MailAccess#connectInternal()
     */
    @Override
    protected void connectInternal() throws OXException {
        //NO-OP
    }

    /* (non-Javadoc)
     * @see com.openexchange.mail.api.MailAccess#createNewMailConfig()
     */
    @Override
    protected MailConfig createNewMailConfig() {
        return new GuardGuestMailConfig();
    }

    /* (non-Javadoc)
     * @see com.openexchange.mail.api.MailAccess#createNewMailProperties()
     */
    @Override
    protected IMailProperties createNewMailProperties() throws OXException {
        return DEFAULT_GUARD_GUEST_MAIL_PROPERTIES;
    }

    /* (non-Javadoc)
     * @see com.openexchange.mail.api.MailAccess#checkMailServerPort()
     */
    @Override
    protected boolean checkMailServerPort() {
        return false; //No mail server port needs to be set
    }

    /* (non-Javadoc)
     * @see com.openexchange.mail.api.MailAccess#releaseResources()
     */
    @Override
    protected void releaseResources() {
    }

    /* (non-Javadoc)
     * @see com.openexchange.mail.api.MailAccess#closeInternal()
     */
    @Override
    protected void closeInternal() {
    }

    /* (non-Javadoc)
     * @see com.openexchange.mail.api.MailAccess#getFolderStorage()
     */
    @Override
    public IGuardGuestFolderStorage getFolderStorage() throws OXException {
        if(folderStorage == null) {
           folderStorage = new GuardGuestFolderStorage(this.session);
        }
        return folderStorage;
    }

    /* (non-Javadoc)
     * @see com.openexchange.mail.api.MailAccess#getMessageStorage()
     */
    @Override
    public IGuardGuestMessageStorage getMessageStorage() throws OXException {
        if(messageStorage == null) {
            messageStorage = new GuardGuestMessageStorage(this.session, GuardGuestFolderStorage.TRASH_NAME);
        }
        return messageStorage;
    }

    /* (non-Javadoc)
     * @see com.openexchange.mail.api.MailAccess#getLogicTools()
     */
    @Override
    public MailLogicTools getLogicTools() throws OXException {
        if(mailLogicTools == null) {
            mailLogicTools = new MailLogicTools(session, accountId);
        }
        return mailLogicTools;
    }

    /* (non-Javadoc)
     * @see com.openexchange.mail.api.MailAccess#isConnected()
     */
    @Override
    public boolean isConnected() {
        return true;
    }

    /* (non-Javadoc)
     * @see com.openexchange.mail.api.MailAccess#isConnectedUnsafe()
     */
    @Override
    public boolean isConnectedUnsafe() {
        // TODO Auto-generated method stub
        return false;
    }

    /* (non-Javadoc)
     * @see com.openexchange.mail.api.MailAccess#startup()
     */
    @Override
    protected void startup() throws OXException {
    }

    /* (non-Javadoc)
     * @see com.openexchange.mail.api.MailAccess#shutdown()
     */
    @Override
    protected void shutdown() throws OXException {
    }

}
